/**********************************************************************************************************************
 * Copyright (c) Prophesee S.A.                                                                                       *
 *                                                                                                                    *
 * Licensed under the Apache License, Version 2.0 (the "License");                                                    *
 * you may not use this file except in compliance with the License.                                                   *
 * You may obtain a copy of the License at http://www.apache.org/licenses/LICENSE-2.0                                 *
 * Unless required by applicable law or agreed to in writing, software distributed under the License is distributed   *
 * on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.                      *
 * See the License for the specific language governing permissions and limitations under the License.                 *
 **********************************************************************************************************************/

#ifndef METAVISION_SDK_DRIVER_ERC_MODULE_H
#define METAVISION_SDK_DRIVER_ERC_MODULE_H

#include <cstdint>
#include <string>

#include "metavision/hal/facilities/i_erc_module.h"

namespace Metavision {

/// @brief Facility class for Event Rate Controller (ERC) commands
class ErcModule {
public:
    /// @brief Constructor
    ErcModule(I_ErcModule *erc);

    /// @brief Destructor
    ~ErcModule();

    /// @brief Toggles ERC activation
    /// @param b Desired state. b = true means ERC active, and b = false means ERC inactive.
    /// @note When ERC is inactive, all the events generated by the sensor are transmitted. Potential
    /// bandwidth limitation might occur.
    /// @return true on success
    bool enable(bool b);

    /// @brief Returns ERC activation state
    /// @return The ERC state
    bool is_enabled();

    /// @brief Sets the target CD event rate of the ERC
    /// @param events_per_sec Event rate expressed in events per second
    /// @note Event rate is computed by ERC block as the number of events generated by the sensor divided by the count
    /// period.
    void set_cd_event_rate(uint32_t events_per_sec);

    /// @brief Gets the CD event rate set point of the ERC
    /// @return The event rate expressed in events per second
    /// @note See @ref set_cd_event_rate to define the event rate target.
    uint32_t get_cd_event_rate();

    /// @brief Gets corresponding facility in HAL library
    I_ErcModule *get_facility() const;

private:
    I_ErcModule *pimpl_;
};

} // namespace Metavision

#endif // METAVISION_SDK_DRIVER_ERC_MODULE_H
